
#include "main.h"
#include "info_bar.h"
#include "map.h"
#include "args.h"


extern int hilight_item;

ThingInfoBar::ThingInfoBar(wxWindow *parent)
:	InfoBar(parent)
{
	wxBoxSizer *hbox = new wxBoxSizer(wxHORIZONTAL);
	SetSizer(hbox);

	// Main frame
	main_frame = new wxStaticBox(this, -1, _T("No Thing Hilighted"));
	wxFont def_font = main_frame->GetFont();
	def_font.SetWeight(wxFONTWEIGHT_BOLD);
	main_frame->SetFont(def_font);
	wxStaticBoxSizer *box = new wxStaticBoxSizer(main_frame, wxHORIZONTAL);
	hbox->Add(box, 1, wxEXPAND|wxRIGHT, 4);

	wxBoxSizer *vbox = new wxBoxSizer(wxVERTICAL);
	vbox->SetMinSize(160, -1);
	box->Add(vbox, 0, wxRIGHT, 8);

	label_type = new wxStaticText(this, -1, _T("Type:"));
	label_position = new wxStaticText(this, -1, _T("Position:"));
	label_angle = new wxStaticText(this, -1, _T("Angle:"));
	label_difficulty = new wxStaticText(this, -1, _T("Difficulty:"));
	label_deaf = new wxStaticText(this, -1, _T("Deaf:"));
	label_multiplayer = new wxStaticText(this, -1, _T("Multiplayer:"));

	vbox->Add(label_type);
	vbox->Add(label_position);
	vbox->Add(label_angle);
	vbox->Add(label_difficulty);
	vbox->Add(label_deaf);
	vbox->Add(label_multiplayer);

	vbox = new wxBoxSizer(wxVERTICAL);
	box->Add(vbox);

	label_arg1 = new wxStaticText(this, -1, _T(""));
	label_arg2 = new wxStaticText(this, -1, _T(""));
	label_arg3 = new wxStaticText(this, -1, _T(""));
	label_arg4 = new wxStaticText(this, -1, _T(""));
	label_arg5 = new wxStaticText(this, -1, _T(""));

	vbox->Add(label_arg1);
	vbox->Add(label_arg2);
	vbox->Add(label_arg3);
	vbox->Add(label_arg4);
	vbox->Add(label_arg5);

	// Sprite frame
	sprite_frame = new wxStaticBox(this, -1, _T("Sprite"));
	sprite_frame->SetFont(def_font);
	box = new wxStaticBoxSizer(sprite_frame, wxVERTICAL);
	hbox->Add(box, 0, wxEXPAND);

	tex_sprite = new TextureBox(this, 96, 96, "_none_", 3, 2.0f);
	box->Add(tex_sprite, 1, wxALIGN_CENTER);

	SetSizeHints(-1, 128, -1, 128);
}

ThingInfoBar::~ThingInfoBar()
{
}

string is_deaf(thing_t *t)
{
	if (t->flags & THING_DEAF)
		return "Yes";
	else
		return "No";
}

string is_multi(thing_t *t)
{
	if (t->flags & THING_MULTI)
		return "Yes";
	else
		return "No";
}

void ThingInfoBar::update()
{
	if (hilight_item == -1)
	{
		main_frame->SetLabel(_T("No Thing Hilighted"));
		label_type->SetLabel(_T("Type:"));
		label_position->SetLabel(_T("Position:"));
		label_angle->SetLabel(_T("Angle:"));
		label_difficulty->SetLabel(_T("Difficulty:"));
		label_deaf->SetLabel(_T("Deaf:"));

		if (map.hexen)
		{
			label_arg1->SetLabel(_T("Arg 1:"));
			label_arg2->SetLabel(_T("Arg 2:"));
			label_arg3->SetLabel(_T("Arg 3:"));
			label_arg4->SetLabel(_T("Arg 4:"));
			label_arg5->SetLabel(_T("Arg 5:"));
			label_multiplayer->SetLabel(_T("TID:"));
		}
		else
		{
			label_arg1->SetLabel(_T(""));
			label_arg2->SetLabel(_T(""));
			label_arg3->SetLabel(_T(""));
			label_arg4->SetLabel(_T(""));
			label_arg5->SetLabel(_T(""));
			label_multiplayer->SetLabel(_T("Multiplayer:"));
		}

		tex_sprite->set_texture("_none_");

		return;
	}

	thing_t *t = map.things[hilight_item];

	main_frame->SetLabel(wxString::Format(_T("Thing #%d: %s"), hilight_item, t->ttype->name.c_str()));
	label_type->SetLabel(wxString::Format(_T("Type: %d"), t->type).c_str());
	if (map.hexen)
		label_position->SetLabel(wxString::Format(_T("Position: (%d, %d, %d)"), t->x, t->y, t->z));
	else
		label_position->SetLabel(wxString::Format(_T("Position: (%d, %d)"), t->x, t->y).c_str());
	label_angle->SetLabel(wxString::Format(_T("Angle: %d (%s)"), t->angle, t->angle_string().c_str()));
	label_difficulty->SetLabel(wxString::Format(_T("Difficulty: %s"), t->difficulty_string().c_str()));
	label_deaf->SetLabel(wxString::Format(_T("Deaf: %s"), is_deaf(t).c_str()));

	if (!map.hexen)
		label_multiplayer->SetLabel(wxString::Format(_T("Multiplayer: %s"), is_multi(t).c_str()));
	else
		label_multiplayer->SetLabel(wxString::Format(_T("TID: %d"), t->tid));

	tex_sprite->set_texture(t->ttype->spritename, 3);

	if (map.hexen)
	{
		thing_type_t* ttype = map.things[hilight_item]->ttype;

		argtype_t *at = get_arg_type(ttype->arg_types[0]);
		if (at)
			label_arg1->SetLabel(wxString::Format(_T("%s: %d (%s)"), ttype->args[0].c_str(), map.things[hilight_item]->args[0], at->get_name(map.things[hilight_item]->args[0]).c_str()));
		else
			label_arg1->SetLabel(wxString::Format(_T("%s: %d"), ttype->args[0].c_str(), map.things[hilight_item]->args[0]));

		at = get_arg_type(ttype->arg_types[1]);
		if (at)
			label_arg2->SetLabel(wxString::Format(_T("%s: %d (%s)"), ttype->args[1].c_str(), map.things[hilight_item]->args[1], at->get_name(map.things[hilight_item]->args[1]).c_str()));
		else
			label_arg2->SetLabel(wxString::Format(_T("%s: %d"), ttype->args[1].c_str(), map.things[hilight_item]->args[1]));

		at = get_arg_type(ttype->arg_types[2]);
		if (at)
			label_arg3->SetLabel(wxString::Format(_T("%s: %d (%s)"), ttype->args[2].c_str(), map.things[hilight_item]->args[2], at->get_name(map.things[hilight_item]->args[2]).c_str()));
		else
			label_arg3->SetLabel(wxString::Format(_T("%s: %d"), ttype->args[2].c_str(), map.things[hilight_item]->args[2]));

		at = get_arg_type(ttype->arg_types[3]);
		if (at)
			label_arg4->SetLabel(wxString::Format(_T("%s: %d (%s)"), ttype->args[3].c_str(), map.things[hilight_item]->args[3], at->get_name(map.things[hilight_item]->args[3]).c_str()));
		else
			label_arg4->SetLabel(wxString::Format(_T("%s: %d"), ttype->args[3].c_str(), map.things[hilight_item]->args[3]));

		at = get_arg_type(ttype->arg_types[4]);
		if (at)
			label_arg5->SetLabel(wxString::Format(_T("%s: %d (%s)"), ttype->args[4].c_str(), map.things[hilight_item]->args[4], at->get_name(map.things[hilight_item]->args[4]).c_str()));
		else
			label_arg5->SetLabel(wxString::Format(_T("%s: %d"), ttype->args[4].c_str(), map.things[hilight_item]->args[4]));
	}

	Layout();
}
