// <<--------------------------------------->>
// << SLADE - SlayeR's 'LeetAss Doom Editor >>
// << By Simon Judd, 2004                   >>
// << ------------------------------------- >>
// << input.cpp - Input stuff               >>
// <<--------------------------------------->>

// Includes ------------------------------- >>
#include "main.h"
#include "keybind.h"
#include "editor_window.h"
#include "edit.h"
#include "edit_misc.h"
#include "map.h"
#include "undoredo.h"
#include "line_edit.h"
#include "thing_edit.h"
#include "sector_edit.h"
#include "struct_3d.h"
#include "mathstuff.h"
#include "camera.h"
#include "3dmode.h"
#include "console.h"
//#include "console_window.h"
#include "copypaste.h"
#include "ttype_select.h"
#include "linedraw.h"

CVAR(Float, move_speed_3d, 0.3f, CVAR_SAVE)
CVAR(Float, mouse_speed_3d, 1.0f, CVAR_SAVE)
CVAR(Int, key_delay_3d, 7, CVAR_SAVE)

#define PRESSED(name) vector_exists(pressed_keys, name)
#define RELEASED(name) vector_exists(released_keys, name)

// External Variables --------------------- >>
extern BindList binds;


extern int edit_mode, hilight_item;
extern float xoff, yoff, zoom;
extern bool /*line_draw,*/paste_mode, /*items_moving,*/ /*thing_quickangle,*/ lock_hilight;
extern point2_t mouse, down_pos;
extern rect_t sel_box;
extern Clipboard clipboard;
extern Camera camera;
extern thing_t last_thing;

extern vector<string> pressed_keys;
extern vector<string> released_keys;

extern EditorWindow *editor_window;

EXTERN_CVAR(Bool, render_fog)
EXTERN_CVAR(Bool, render_fullbright)
EXTERN_CVAR(Bool, render_hilight)
EXTERN_CVAR(Int, render_things)
EXTERN_CVAR(Bool, edit_snap_grid)

// cycle_edit_mode: Cycles the edit mode
// ----------------------------------- >>
void cycle_edit_mode()
{
	int nmode = edit_mode;

	if (++nmode > 3)
		nmode = 0;

	change_edit_mode(nmode);
	editor_window->update_statusbar();
}

// edit_item: When the edit item key is pressed
// ----------------------------------------- >>
void edit_item()
{
	if (edit_mode == 0)
	{
		create_vertex();
		get_hilight_item(mouse.x, mouse.y);
		editor_window->redraw_map(true, false);
		return;
	}

	if (hilight_item == -1 && !selection())
		return;

	//wxLogMessage("Edit Item");
	bool old_lock = lock_hilight;
	lock_hilight = true;

	if (edit_mode == 1)
	{
		LineEditor le;
		if (le.ShowModal() == wxID_OK)
			le.apply_changes();
	}

	if (edit_mode == 3)
	{
		ThingEditor te;
		if (te.ShowModal() == wxID_OK)
			te.apply_changes();
	}

	if (edit_mode == 2)
	{
		SectorEditor se;
		if (se.ShowModal() == wxID_OK)
			se.apply_changes();
	}

	editor_window->redraw_map(true, false);
	lock_hilight = old_lock;
}

void view_zoom(bool in)
{
	if (in)
	{
		zoom *= 1.2f;

		if (zoom > 1024)
			zoom = 1024;
	}
	else
	{
		zoom *= 0.8f;

		if (zoom < 0.001f)
			zoom = 0.001f;
	}

	editor_window->redraw_map(true, true);
}

// keys_edit: Keys for the 2d editor
// ------------------------------ >>
void keys_edit()
{
	//wxLogMessage("keys_edit()");

	if (!map.opened)
		return;

	// **** Dragging stuff ****


	// Selection box
	if (sel_box.x1() != -1 && RELEASED("edit_selectbox"))
	{
		select_items_box(sel_box);
		editor_window->redraw_map(true, false);
		sel_box.set(-1, -1, -1, -1);
		return;
	}

	// Moving items
	if (state(STATE_MOVING) && RELEASED("edit_moveitems"))
	{
		clear_move_items();
		editor_window->redraw_map(true, false);
		change_state();
		return;
	}

	// Quick thing angle
	if (state(STATE_THINGANGLE) && RELEASED("thing_quickangle"))
	{
		//thing_quickangle = false;
		change_state();
		editor_window->redraw_map(true, false);
		return;
	}

	// Map panning
	if (state(STATE_MAPPAN) && RELEASED("view_panmap"))
	{
		change_state();
		editor_window->redraw_map(true, true);
		return;
	}

	// Moving items
	if (state(STATE_MOVING) || sel_box.x1() != -1)
		return;

	// **** Other controls ****

	// Scroll up
	if (PRESSED("view_up"))
	{
		yoff -= ((1.0f / zoom) * 64);
		//yoff += ((MAJOR_UNIT / (int)zoom)) + 1;
		editor_window->redraw_map(true, true);
	}
	
	// Scroll down
	if (PRESSED("view_down"))
	{
		yoff += ((1.0f / zoom) * 64);
		//yoff -= ((MAJOR_UNIT / (int)zoom)) + 1;
		editor_window->redraw_map(true, true);
	}

	// Scroll left
	if (PRESSED("view_left"))
	{
		xoff -= ((1.0f / zoom) * 64);
		//xoff += ((MAJOR_UNIT / (int)zoom)) + 1;
		editor_window->redraw_map(true, true);
	}
	
	// Scroll right
	if (PRESSED("view_right"))
	{
		xoff += ((1.0f / zoom) * 64);
		//xoff -= ((MAJOR_UNIT / (int)zoom)) + 1;
		editor_window->redraw_map(true, true);
	}

	// Zoom in
	if (PRESSED("view_zoomin"))
		view_zoom(true);

	// Zoom out
	if (PRESSED("view_zoomout"))
		view_zoom(false);

	// Center view on mouse
	if (RELEASED("view_mousecenter"))
	{
		xoff = m_x(mouse.x);
		yoff = m_y(mouse.y);
		editor_window->redraw_map(true, true);
	}

	// Set offsets to 0, 0
	if (RELEASED("view_origin"))
	{
		xoff = yoff = 0;
		editor_window->redraw_map(true, true);
	}

	// Popup context menu
	if (RELEASED("view_contextmenu") && state())
	{
		binds.clear("view_contextmenu");
		editor_window->popup_context_menu();
	}

	// Sidebar stuff
	wxCommandEvent event;
	if (RELEASED("view_hidesidebar"))
		editor_window->view_hidesidebar(event);

	if (RELEASED("view_drawsidebar"))
		editor_window->view_drawsidebar(event);

	if (RELEASED("view_themesidebar"))
		editor_window->view_themesidebar(event);

	// Vertices mode
	if (RELEASED("mode_vertices"))
		change_edit_mode(0);
	
	// Linedefs mode
	if (RELEASED("mode_linedefs"))
		change_edit_mode(1);
	
	// Sectors mode
	if (RELEASED("mode_sectors"))
		change_edit_mode(2);
	
	// Things mode
	if (RELEASED("mode_things"))
		change_edit_mode(3);

	// Change mode
	if (RELEASED("mode_change"))
		cycle_edit_mode();

	// Increase grid size
	if (RELEASED("view_increasegrid"))
	{
		increase_grid();
		editor_window->redraw_map(false, true);
	}

	// Decrease grid size
	if (RELEASED("view_decreasegrid"))
	{
		decrease_grid();
		editor_window->redraw_map(false, true);
	}

	// Clear selection
	if (RELEASED("edit_clearselection") && state())
	{
		clear_selection();
		editor_window->redraw_map(true);
	}

	// Delete item
	if (RELEASED("edit_deleteitem") && state())
	{
		if (edit_mode == 0)
			delete_vertex();

		if (edit_mode == 1)
			delete_line();

		if (edit_mode == 2)
			delete_sector();

		if (edit_mode == 3)
			delete_thing();

		editor_window->redraw_map(true);
		lock_hilight = false;
	}

	if (PRESSED("edit_deleteitem") && state())
		lock_hilight = true;

	// Create item
	if (RELEASED("edit_createitem") && state())
	{
		if (edit_mode == 0)
		{
			if (!selection())
				create_vertex();
			else
				create_lines(false);

			editor_window->redraw_map(true);
			return;
		}

		if (edit_mode == 1)
		{
			if (selection())
				create_sector();

			editor_window->redraw_map(true);
			return;
		}

		if (edit_mode == 3)
		{
			create_thing();
			editor_window->redraw_map(true);
			return;
		}

		get_hilight_item(mouse.x, mouse.y);
		binds.clear("edit_createitem");
	}

	// Create item 2
	if (RELEASED("edit_createitem2") && state())
	{
		if (edit_mode == 0)
		{
			if (!selection())
				create_vertex();
			else
				create_lines(true);

			editor_window->redraw_map(true);
			return;
		}

		if (edit_mode == 1)
		{
			if (selection())
				create_sector();

			editor_window->redraw_map(true);
			return;
		}

		if (edit_mode == 3)
		{
			TTypeSelectDialog td(last_thing.type);

			if (td.ShowModal() == wxID_OK)
			{
				last_thing.type = td.get_type();
				create_thing();
			}

			editor_window->redraw_map(true);
			return;
		}

		get_hilight_item(mouse.x, mouse.y);
		binds.clear("edit_createitem2");
	}

	// Select item
	if (RELEASED("edit_selectitem") && state())
	{
		select_item();
		editor_window->redraw_map(true);
	}

	// Sector height quick changes (8 units)
	if (PRESSED("sector_upfloor8"))
	{
		if (edit_mode == 2)
			sector_changeheight(true, 8);
	}

	if (PRESSED("sector_downfloor8"))
	{
		if (edit_mode == 2)
			sector_changeheight(true, -8);
	}

	if (PRESSED("sector_upceil8"))
	{
		if (edit_mode == 2)
			sector_changeheight(false, 8);
	}

	if (PRESSED("sector_downceil8"))
	{
		if (edit_mode == 2)
			sector_changeheight(false, -8);
	}

	if (PRESSED("sector_upboth8"))
	{
		if (edit_mode == 2)
		{
			sector_changeheight(true, 8);
			sector_changeheight(false, 8);
		}
	}

	if (PRESSED("sector_downboth8"))
	{
		if (edit_mode == 2)
		{
			sector_changeheight(true, -8);
			sector_changeheight(false, -8);
		}
	}

	// Sector height quick changes (1 unit)
	if (PRESSED("sector_upfloor"))
	{
		if (edit_mode == 2)
			sector_changeheight(true, 1);
	}

	if (PRESSED("sector_downfloor"))
	{
		if (edit_mode == 2)
			sector_changeheight(true, -1);
	}

	if (PRESSED("sector_upceil"))
	{
		if (edit_mode == 2)
			sector_changeheight(false, 1);
	}

	if (PRESSED("sector_downceil"))
	{
		if (edit_mode == 2)
			sector_changeheight(false, -1);
	}

	if (PRESSED("sector_upboth"))
	{
		if (edit_mode == 2)
		{
			sector_changeheight(true, 1);
			sector_changeheight(false, 1);
		}
	}

	if (PRESSED("sector_downboth"))
	{
		if (edit_mode == 2)
		{
			sector_changeheight(true, -1);
			sector_changeheight(false, -1);
		}
	}

	// Sector light change
	if (PRESSED("sector_uplight"))
	{
		//binds.clear("sector_uplight");
		sector_changelight(16);
	}

	if (PRESSED("sector_downlight"))
	{
		//binds.clear("sector_downlight");
		sector_changelight(-16);
	}

	// Flip line
	if (RELEASED("line_flip") && state())
	{
		if (edit_mode == 1)
			line_flip(true, false);

		editor_window->redraw_map(true);
	}

	// Swap line sides
	if (RELEASED("line_swapsides") && state())
	{
		if (edit_mode == 1)
			line_flip(false, true);

		editor_window->redraw_map(true);
	}

	// Flip both line direction and sides
	if (RELEASED("line_flipboth") && state())
	{
		if (edit_mode == 1)
			line_flip(true, true);

		editor_window->redraw_map(true);
	}

	if (RELEASED("line_correctrefs") && state())
		line_correct_references();

	// Begin line draw
	if (PRESSED("line_begindraw"))
	{
		//if (!line_draw)
		//	line_draw = true;

		change_state(STATE_LINEDRAW);

		binds.clear("line_begindraw");
	}

	// Begin rectangle draw
	if (PRESSED("line_begindraw_rect"))
	{
		/*
		if (!line_draw)
		{
			line_draw = true;
			sel_box.set(mouse.x, mouse.y, mouse.x, mouse.y);
		}
		*/

		change_state(STATE_LINEDRAW);
		sel_box.set(mouse.x, mouse.y, mouse.x, mouse.y);

		binds.clear("line_begindraw_rect");
	}

	// Undo
	if (RELEASED("edit_undo") && state())
	{
		undo();
		clear_selection();
		hilight_item = -1;
		editor_window->redraw_map(true, true);
		//map_changelevel(3);
		map.change_level(MC_NODE_REBUILD);
		binds.clear("edit_undo");
	}

	if (RELEASED("edit_createsector"))
	{
		sector_create(point2_t(m_x(mouse.x), -m_y(mouse.y)));
	}

	// Edit item
	if (RELEASED("edit_edititem") && state())
	{
		edit_item();
		binds.clear("edit_edititem");
	}

	// Merge sectors
	if (RELEASED("sector_merge") && state())
	{
		sector_merge(false);
		binds.clear("sector_merge");
	}

	// Join sectors
	if (RELEASED("sector_join") && state())
	{
		sector_merge(true);
		binds.clear("sector_join");
	}

	if (RELEASED("view_3dmode") && state())
	{
		binds.clear("view_3dmode");
		binds.clear("3d_exit");
		start_3d_mode();
	}

	if (RELEASED("open_console"))
	{
		binds.clear("open_console");
		//popup_console();
	}

	if (RELEASED("copy"))
	{
		binds.clear("copy");
		clipboard.Copy();
	}

	if (RELEASED("paste") && state())
	{
		binds.clear("paste");
		//paste_mode = true;
		change_state(STATE_PASTE);
		clear_selection();
	}

	if (RELEASED("cancel_paste"))
	{
		binds.clear("cancel_paste");
		//paste_mode = false;
		change_state();
		editor_window->redraw_map(true, false);
	}

	// Toggle grid snap
	if (RELEASED("edit_gridsnap"))
	{
		binds.clear("edit_gridsnap");
		edit_snap_grid = !edit_snap_grid;
		editor_window->redraw_map(false, false);
	}

	// Toggle hilight lock
	if (RELEASED("edit_lockhilight"))
	{
		binds.clear("edit_lockhilight");
		lock_hilight = !lock_hilight;
	}
}

void keys_linedraw()
{
	// Scroll up
	if (PRESSED("view_up"))
	{
		yoff -= ((1.0f / zoom) * 64);
		//yoff += ((MAJOR_UNIT / (int)zoom)) + 1;
		editor_window->redraw_map(true, true);
	}
	
	// Scroll down
	if (PRESSED("view_down"))
	{
		yoff += ((1.0f / zoom) * 64);
		//yoff -= ((MAJOR_UNIT / (int)zoom)) + 1;
		editor_window->redraw_map(true, true);
	}

	// Scroll left
	if (PRESSED("view_left"))
	{
		xoff -= ((1.0f / zoom) * 64);
		//xoff += ((MAJOR_UNIT / (int)zoom)) + 1;
		editor_window->redraw_map(true, true);
	}
	
	// Scroll right
	if (PRESSED("view_right"))
	{
		xoff += ((1.0f / zoom) * 64);
		//xoff -= ((MAJOR_UNIT / (int)zoom)) + 1;
		editor_window->redraw_map(true, true);
	}

	// Zoom in
	if (PRESSED("view_zoomin"))
		view_zoom(true);

	// Zoom out
	if (PRESSED("view_zoomout"))
		view_zoom(false);

	// Increase grid size
	if (RELEASED("view_increasegrid"))
	{
		increase_grid();
		editor_window->redraw_map(false, true);
	}

	// Decrease grid size
	if (RELEASED("view_decreasegrid"))
	{
		decrease_grid();
		editor_window->redraw_map(false, true);
	}

	// Toggle grid snap
	if (RELEASED("edit_gridsnap"))
	{
		binds.clear("edit_gridsnap");
		edit_snap_grid = !edit_snap_grid;
		editor_window->redraw_map(false, false);
	}

	// Draw point
	if (RELEASED("ldraw_drawpoint"))
	{
		line_drawpoint(false);
		editor_window->redraw_map(false, false);
	}

	// Undraw point
	if (RELEASED("ldraw_cancelpoint"))
	{
		line_undrawpoint();
		editor_window->redraw_map(false, false);
	}

	// Draw point at nearest vertex
	if (RELEASED("ldraw_nearestvert"))
	{
		line_drawpoint(true);
		editor_window->redraw_map(false, false);
	}
}

//#define KEY_3D_DELAY 7
int key_3d_rep = 0;

bool keys_3d(bool mwheel)
{
	bool key_3d_allow = false;
	float speed = 0.2 * move_speed_3d;

	if (key_3d_rep == 0)
		key_3d_allow = true;

	if (binds.pressed("3d_exit"))
	{
		binds.clear("view_3dmode");
		binds.clear("3d_exit");
		return false;
	}

	if (binds.pressed("3d_forward"))
		camera.move_camera(speed);

	if (binds.pressed("3d_back"))
		camera.move_camera(-speed);

	if (binds.pressed("3d_left"))
		camera.rotate_view(speed, 0.0f, 0.0f, 0.5f);

	if (binds.pressed("3d_right"))
		camera.rotate_view(-speed, 0.0f, 0.0f, 0.5f);

	if (binds.pressed("3d_strafeleft"))
		camera.strafe_camera(-speed);

	if (binds.pressed("3d_straferight"))
		camera.strafe_camera(speed);

	if (binds.pressed("3d_moveup"))
	{
		camera.position.z += 0.1f;
		camera.view.z += 0.1f;
	}

	if (binds.pressed("3d_movedown"))
	{
		camera.position.z -= 0.1f;
		camera.view.z -= 0.1f;
	}

	if (binds.pressed("3d_toggle_fullbright"))
	{
		binds.clear("3d_toggle_fullbright");
		render_fullbright = !render_fullbright;
	}

	if (binds.pressed("3d_toggle_fog"))
	{
		binds.clear("3d_toggle_fog");
		render_fog = !render_fog;
	}
	
	if (binds.pressed("3d_toggle_things"))
	{
		binds.clear("3d_toggle_things");
		render_things = render_things + 1;

		if (render_things > 3)
			render_things = 0;
	}

	if (binds.pressed("3d_toggle_gravity"))
	{
		binds.clear("3d_toggle_gravity");
		camera.gravity = !camera.gravity;

		if (camera.gravity)
			add_3d_message("Gravity ON");
		else
			add_3d_message("Gravity OFF");
	}

	// Sector height quick changes (8 units)
	if (binds.pressed("3d_upfloor8") && (key_3d_allow || mwheel))
	{
		change_sector_height_3d(8);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_downfloor8") && (key_3d_allow || mwheel))
	{
		change_sector_height_3d(-8);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_upceil8") && (key_3d_allow || mwheel))
	{
		change_sector_height_3d(8, false);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_downceil8") && (key_3d_allow || mwheel))
	{
		change_sector_height_3d(-8, false);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_upboth8") && (key_3d_allow || mwheel))
	{
		change_sector_height_3d(8, true, true);
		change_sector_height_3d(8, false, true);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_downboth8") && (key_3d_allow || mwheel))
	{
		change_sector_height_3d(-8, true, true);
		change_sector_height_3d(-8, false, true);
		key_3d_rep = key_delay_3d;
	}

	// Sector height quick changes (1 unit)
	if (binds.pressed("3d_upfloor") && (key_3d_allow || mwheel))
	{
		change_sector_height_3d(1);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_downfloor") && (key_3d_allow || mwheel))
	{
		change_sector_height_3d(-1);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_upceil") && (key_3d_allow || mwheel))
	{
		change_sector_height_3d(1, false);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_downceil") && (key_3d_allow || mwheel))
	{
		change_sector_height_3d(-1, false);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_upboth") && (key_3d_allow || mwheel))
	{
		change_sector_height_3d(1, true, true);
		change_sector_height_3d(1, false, true);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_downboth") && (key_3d_allow || mwheel))
	{
		change_sector_height_3d(-1, true, true);
		change_sector_height_3d(-1, false, true);
		key_3d_rep = key_delay_3d;
	}

	// Texture offset
	if (binds.pressed("3d_upyoffset") && (key_3d_allow || mwheel))
	{
		change_offsets_3d(0, 1);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_downyoffset") && (key_3d_allow || mwheel))
	{
		change_offsets_3d(0, -1);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_upxoffset") && (key_3d_allow || mwheel))
	{
		change_offsets_3d(1, 0);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_downxoffset") && (key_3d_allow || mwheel))
	{
		change_offsets_3d(-1, 0);
		key_3d_rep = key_delay_3d;
	}

	// Texture offset x8
	if (binds.pressed("3d_upyoffset8") && (key_3d_allow || mwheel))
	{
		change_offsets_3d(0, 8);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_downyoffset8") && (key_3d_allow || mwheel))
	{
		change_offsets_3d(0, -8);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_upxoffset8") && (key_3d_allow || mwheel))
	{
		change_offsets_3d(8, 0);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_downxoffset8") && (key_3d_allow || mwheel))
	{
		change_offsets_3d(-8, 0);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_toggle_hilight"))
	{
		binds.clear("3d_toggle_hilight");
		render_hilight = !render_hilight;
	}

	// Upper/Lower unpegged toggle
	if (binds.pressed("3d_upperunpegged"))
	{
		binds.clear("3d_upperunpegged");
		toggle_texture_peg_3d(true);
	}

	if (binds.pressed("3d_lowerunpegged"))
	{
		binds.clear("3d_lowerunpegged");
		toggle_texture_peg_3d(false);
	}

	// Change light level
	if (binds.pressed("3d_uplightlevel"))
	{
		binds.clear("3d_uplightlevel");
		change_light_3d(16);
	}

	if (binds.pressed("3d_downlightlevel"))
	{
		binds.clear("3d_downlightlevel");
		change_light_3d(-16);
	}

	if (binds.pressed("3d_uplightlevel1"))
	{
		binds.clear("3d_uplightlevel1");
		change_light_3d(1);
	}

	if (binds.pressed("3d_downlightlevel1"))
	{
		binds.clear("3d_downlightlevel1");
		change_light_3d(-1);
	}

	// Change thing angle
	if (binds.pressed("3d_upthingangle") && (key_3d_allow || mwheel))
	{
		change_thing_angle_3d(45);
		key_3d_rep = key_delay_3d;
	}

	if (binds.pressed("3d_downthingangle") && (key_3d_allow || mwheel))
	{
		change_thing_angle_3d(-45);
		key_3d_rep = key_delay_3d;
	}

	// Change thing z height
	if (map.hexen)
	{
		if (binds.pressed("3d_upthingz8") && (key_3d_allow || mwheel))
		{
			change_thing_z_3d(8);
			key_3d_rep = key_delay_3d;
		}

		if (binds.pressed("3d_downthingz8") && (key_3d_allow || mwheel))
		{
			change_thing_z_3d(-8);
			key_3d_rep = key_delay_3d;
		}

		if (binds.pressed("3d_upthingz") && (key_3d_allow || mwheel))
		{
			change_thing_z_3d(1);
			key_3d_rep = key_delay_3d;
		}

		if (binds.pressed("3d_downthingz") && (key_3d_allow || mwheel))
		{
			change_thing_z_3d(-1);
			key_3d_rep = key_delay_3d;
		}
	}

	if (binds.pressed("3d_align_tex_x"))
	{
		binds.clear("3d_align_tex_x");
		auto_align_x_3d();
	}

	// Reset offsets
	if (binds.pressed("3d_reset_offsets"))
	{
		binds.clear("3d_reset_offsets");
		reset_offsets_3d();
	}

	if (binds.pressed("3d_reset_xoffset"))
	{
		binds.clear("3d_reset_xoffset");
		reset_offsets_3d(true, false);
	}

	if (binds.pressed("3d_reset_yoffset"))
	{
		binds.clear("3d_reset_yoffset");
		reset_offsets_3d(false, true);
	}

	// Change (browse) texture
	if (binds.pressed("3d_change_texture"))
	{
		binds.clear("3d_change_texture");
		change_texture_3d(false);
	}

	if (binds.pressed("3d_change_tex_paint"))
	{
		binds.clear("3d_change_tex_paint");
		change_texture_3d(true);
	}

	// Copy/Paste texture
	if (binds.pressed("3d_copy_texture"))
	{
		binds.clear("3d_copy_texture");
		copy_texture_3d();
	}

	if (binds.pressed("3d_paste_texture"))
	{
		binds.clear("3d_paste_texture");
		paste_texture_3d(false);
	}

	if (binds.pressed("3d_paste_paint"))
	{
		binds.clear("3d_paste_paint");
		paste_texture_3d(true);
	}

	// Copy/paste side properties
	if (binds.pressed("3d_copy_side"))
	{
		binds.clear("3d_copy_side");
		copy_side_3d(true, true, true);
	}

	if (binds.pressed("3d_paste_side"))
	{
		binds.clear("3d_paste_side");
		paste_side_3d(true, true, true);
	}

	// Copy/paste offsets
	if (binds.pressed("3d_copy_offsets"))
	{
		binds.clear("3d_copy_offsets");
		copy_side_3d(true, true);
	}

	if (binds.pressed("3d_paste_offsets"))
	{
		binds.clear("3d_paste_offsets");
		paste_side_3d(true, true);
	}

	if (binds.pressed("3d_copy_xoffset"))
	{
		binds.clear("3d_copy_xoffset");
		copy_side_3d(true, false);
	}

	if (binds.pressed("3d_paste_xoffset"))
	{
		binds.clear("3d_paste_xoffset");
		paste_side_3d(true, false);
	}

	if (binds.pressed("3d_copy_yoffset"))
	{
		binds.clear("3d_copy_yoffset");
		copy_side_3d(false, true);
	}

	if (binds.pressed("3d_paste_yoffset"))
	{
		binds.clear("3d_paste_yoffset");
		paste_side_3d(false, true);
	}

	key_3d_rep--;
	if (key_3d_rep < 0)
		key_3d_rep = 0;

	return true;
}
