
#include "textures.h"

#define SCALE_3D	0.01f

#define PART_UPPER	0	// Upper 'texture'
#define PART_MIDDLE	1	// Middle (solid, 1-sided) 'texture'
#define PART_LOWER	2	// Lower 'texture'
#define PART_TRANS	3	// Transparent middle (bars, etc)

#define PART_FLOOR	0
#define PART_CEIL	1

// Wallrect flags
#define WRECT_SKY	0x01

// Flatpoly flags
#define FPOLY_SKY	0x01


// WallRect flags
#define WR_UPPER		0x01
#define WR_LOWER		0x02
#define WR_MIDDLE		0x03
#define WR_TRANS		0x04	// When set, wall is translucent/transparent (draw after normal walls)
#define WR_SIDE			0x08	// 0=First side, 1=Second side
#define WR_TOP_SWAP		0x10	// When set, wrect uses the floor height of top_sector for the top
#define WR_BOTTOM_SWAP	0x20	// When set, wrect uses the ceiling height of bottom_sector for the bottom
#define WR_TEX_SETUP	0x40	// When not set, setup texture coordinates next draw
#define WR_SKY			0x80	// When set, wrect is sky (fullbright/sky texture)

class WallRect
{
private:
	//BYTE		part;
	BYTE		flags;
	int			top_sector;		// Sector floor/ceil along the top of the rect
	int			bottom_sector;	// Sector floor/ceil along the bottom of the rect
	int			prop_sector;	// Sector to get light/colour etc properties from
	int			geom_line;		// Line to get geometry info from (start & end verts)
	int			prop_line;		// Line to get texture/alignment properties from

public:
	WallRect();
	~WallRect(){}

	int g_line() { return geom_line; }

	void set_flag(BYTE flag, bool set = true)
	{
		if (set)
			flags |= flag;
		else
			flags = (flags & ~flag);
	}

	void set_flags(BYTE f) { flags = f; }
	void set_tsect(int s) { top_sector = s; }
	void set_bsect(int s) { bottom_sector = s; }
	void set_psect(int s) { prop_sector = s; }
	void set_gline(int l) { geom_line = l; }
	void set_pline(int l) { prop_line = l; }

	void draw();
};

struct wallrect_t
{
	// Parent info
	int					sector;
	int					line;
	BYTE				side;
	BYTE				part;

	// Draw info
	Texture*			tex;
	vector<point3_t>	verts;
	BYTE				flags;

	wallrect_t();
	~wallrect_t();

	void draw();
	void draw_hilight();
};

struct flatpoly_t
{
	int		parent_sector;
	BYTE	part;			// 0 = floor, 1 = ceiling

	// Draw info
	//GLuint				tex;
	Texture*			tex;
	vector<point3_t>	verts;
	vector<bool>		lines;
	BYTE				flags;

	flatpoly_t();
	~flatpoly_t();

	void draw();
	void draw_hilight();
};

struct thing3d_t
{
	Texture*	sprite;
	int			parent_sector;

	thing3d_t();
};

struct sectinfo_t
{
	plane_t	f_plane;
	plane_t	c_plane;
	bool visible;
};

struct line3d_t
{
	vector<wallrect_t*> rects;
	bool				visible;
	BYTE				alpha;
};

struct ssect3d_t
{
	vector<flatpoly_t*> flats;
	bool visible;
};

bool setup_wallrect(wallrect_t *wall);
void setup_flatpoly(flatpoly_t *poly, int ssector);
void setup_sector(int s);
void setup_3d_line(int line);
void render_3d_view(int width, int height);
bool setup_3d_data();
void setup_specials_3d(int sector = -1);
void setup_slopes_3d(int sector = -1);
float plane_height(plane_t plane, float x, float y);
