
#include "main.h"
#include "sector_edit.h"
#include "map.h"
#include "checks.h"
#include "stype_select.h"


extern int hilight_item;
extern vector<int> selected_items;

SectorEditor::SectorEditor()
:	wxDialog(NULL, -1, _T("Edit Sector"), wxDefaultPosition, wxDefaultSize)
{
	wxBoxSizer *m_vbox = new wxBoxSizer(wxVERTICAL);
	SetSizer(m_vbox);

	wxBoxSizer *hbox = new wxBoxSizer(wxHORIZONTAL);
	m_vbox->Add(hbox, 0, wxEXPAND);

	// Floor frame
	wxStaticBox *frame = new wxStaticBox(this, -1, _T("Floor"));
	wxStaticBoxSizer *box = new wxStaticBoxSizer(frame, wxVERTICAL);
	hbox->Add(box, 0, wxEXPAND|wxALL, 4);

	entry_ftex = new wxTextCtrl(this, SE_ENTRY_FTEX, _T(""), wxDefaultPosition, wxDefaultSize, wxTE_CENTRE);
	tbox_ftex = new ClickableTexBox(this, 128, 128, entry_ftex, "_none_", 2, 5.0f);
	tbox_ftex->Enable(true);
	box->Add(tbox_ftex, 0, wxEXPAND|wxALL, 4);
	box->Add(entry_ftex, 0, wxEXPAND|wxALL, 4);

	wxBoxSizer *hbox2 = new wxBoxSizer(wxHORIZONTAL);
	box->Add(hbox2, 0, wxEXPAND);

	entry_fheight = new wxTextCtrl(this, -1, _T(""), wxDefaultPosition, wxSize(48, -1));
	hbox2->Add(new wxStaticText(this, -1, _T("Height:")), 1, wxALIGN_CENTER_VERTICAL);
	hbox2->Add(entry_fheight, 0, wxEXPAND|wxALL, 4);

	// Ceiling frame
	frame = new wxStaticBox(this, -1, _T("Ceiling"));
	box = new wxStaticBoxSizer(frame, wxVERTICAL);
	hbox->Add(box, 0, wxEXPAND|wxALL, 4);

	entry_ctex = new wxTextCtrl(this, SE_ENTRY_CTEX, _T(""), wxDefaultPosition, wxDefaultSize, wxTE_CENTRE);
	tbox_ctex = new ClickableTexBox(this, 128, 128, entry_ctex, "_none_", 2, 5.0f);
	tbox_ctex->Enable(true);
	box->Add(tbox_ctex, 0, wxEXPAND|wxALL, 4);
	box->Add(entry_ctex, 0, wxEXPAND|wxALL, 4);

	hbox2 = new wxBoxSizer(wxHORIZONTAL);
	box->Add(hbox2, 0, wxEXPAND);

	entry_cheight = new wxTextCtrl(this, -1, _T(""), wxDefaultPosition, wxSize(48, -1));
	hbox2->Add(new wxStaticText(this, -1, _T("Height:")), 1, wxALIGN_CENTER_VERTICAL);
	hbox2->Add(entry_cheight, 0, wxEXPAND|wxALL, 4);

	// Properties frame
	frame = new wxStaticBox(this, -1, _T("Properties"));
	box = new wxStaticBoxSizer(frame, wxVERTICAL);
	m_vbox->Add(box, 0, wxEXPAND|wxALL, 4);

	// Special
	hbox = new wxBoxSizer(wxHORIZONTAL);
	box->Add(hbox, 0, wxEXPAND|wxLEFT|wxRIGHT|wxTOP, 4);

	entry_special = new wxTextCtrl(this, SE_ENTRY_SPECIAL, _T(""), wxDefaultPosition, wxSize(48, -1));
	btn_changespecial = new wxButton(this, SE_BTN_CHANGESPECIAL, _T("Change"));
	hbox->Add(new wxStaticText(this, -1, _T("Special:")), 1, wxALIGN_CENTER_VERTICAL);
	hbox->Add(entry_special, 0, wxRIGHT|wxALIGN_CENTER_VERTICAL, 4);
	hbox->Add(btn_changespecial, 0, wxEXPAND);

	label_special = new wxStaticText(this, -1, _T(""));
	box->Add(label_special, 0, wxEXPAND|wxLEFT|wxRIGHT, 4);

	// Tag
	hbox = new wxBoxSizer(wxHORIZONTAL);
	box->Add(hbox, 0, wxEXPAND|wxLEFT|wxRIGHT|wxBOTTOM, 4);

	entry_tag = new wxTextCtrl(this, -1, _T(""), wxDefaultPosition, wxSize(48, -1));
	btn_findtag = new wxButton(this, SE_BTN_FINDTAG, _T("Free Tag"));
	hbox->Add(new wxStaticText(this, -1, _T("Sector Tag:")), 1, wxALIGN_CENTER_VERTICAL);
	hbox->Add(entry_tag, 0, wxRIGHT|wxALIGN_CENTER_VERTICAL, 4);
	hbox->Add(btn_findtag, 0, wxEXPAND);

	// Light
	hbox = new wxBoxSizer(wxHORIZONTAL);
	box->Add(hbox, 0, wxEXPAND|wxALL, 4);

	entry_light = new wxTextCtrl(this, -1, _T(""), wxDefaultPosition, wxSize(48, -1));
	hbox->Add(new wxStaticText(this, -1, _T("Light Level:")), 1, wxALIGN_CENTER_VERTICAL);
	hbox->Add(entry_light, 0);

	m_vbox->Add(CreateButtonSizer(wxOK|wxCANCEL), 0, wxEXPAND|wxALL, 4);

	setup_widgets();
	Layout();
	SetBestFittingSize();
	Center();
}

SectorEditor::~SectorEditor()
{
}

void SectorEditor::setup_widgets()
{
	sector_t* s;

	if (selected_items.size() == 0 && hilight_item != -1)
		s = map.sectors[hilight_item];
	else if (selected_items.size() > 0)
		s = map.sectors[selected_items[0]];
	else
		return;

	// Floor
	tbox_ftex->set_texture(s->f_tex, 2);
	entry_ftex->SetValue(str_to_wx(s->f_tex));
	entry_fheight->SetValue(wxString::Format(_T("%d"), s->f_height));

	// Ceiling
	tbox_ctex->set_texture(s->c_tex, 2);
	entry_ctex->SetValue(str_to_wx(s->c_tex));
	entry_cheight->SetValue(wxString::Format(_T("%d"), s->c_height));

	// Special
	entry_special->SetValue(wxString::Format(_T("%d"), s->special));

	// Tag
	entry_tag->SetValue(wxString::Format(_T("%d"), s->tag));

	// Light
	entry_light->SetValue(wxString::Format(_T("%d"), s->light));

	// Multiple items, check for inconsistency
	if (selected_items.size() > 0)
	{
		for (int a = 0; a < selected_items.size(); a++)
		{
			// Floor tex
			if (map.sectors[selected_items[a]]->f_tex != s->f_tex)
			{
				tbox_ftex->set_texture("_none_", 2);
				entry_ftex->SetValue(_T(""));
			}

			// Ceiling tex
			if (map.sectors[selected_items[a]]->c_tex != s->c_tex)
			{
				tbox_ctex->set_texture("_none_", 2);
				entry_ctex->SetValue(_T(""));
			}

			// Floor height
			if (map.sectors[selected_items[a]]->f_height != s->f_height)
				entry_fheight->SetValue(_T(""));

			// Ceiling height
			if (map.sectors[selected_items[a]]->c_height != s->c_height)
				entry_cheight->SetValue(_T(""));

			// Special
			if (map.sectors[selected_items[a]]->special != s->special)
				entry_special->SetValue(_T(""));

			// Tag
			if (map.sectors[selected_items[a]]->tag != s->tag)
				entry_tag->SetValue(_T(""));

			// Light
			if (map.sectors[selected_items[a]]->light != s->light)
				entry_light->SetValue(_T(""));
		}
	}
}

void SectorEditor::apply_changes()
{
	vector<sector_t*> items;

	// Get items to modify
	if (selected_items.size() == 0 && hilight_item != -1)
		items.push_back(map.sectors[hilight_item]);

	if (selected_items.size() > 0)
	{
		for (int a = 0; a < selected_items.size(); a++)
			items.push_back(map.sectors[selected_items[a]]);
	}

	for (int a = 0; a < items.size(); a++)
	{
		// Floor tex
		if (entry_ftex->GetValue() != _T(""))
			items[a]->f_tex = wx_to_str(entry_ftex->GetValue().Upper());

		// Floor height
		if (entry_fheight->GetValue() != _T(""))
			items[a]->f_height = atoi(wx_to_str(entry_fheight->GetValue()).c_str());

		// Ceiling tex
		if (entry_ctex->GetValue() != _T(""))
			items[a]->c_tex = wx_to_str(entry_ctex->GetValue().Upper());

		// Ceiling height
		if (entry_cheight->GetValue() != _T(""))
			items[a]->c_height = atoi(wx_to_str(entry_cheight->GetValue()).c_str());

		// Special
		if (entry_special->GetValue() != _T(""))
			items[a]->special = atoi(wx_to_str(entry_special->GetValue()).c_str());

		// Tag
		if (entry_tag->GetValue() != _T(""))
			items[a]->tag = atoi(wx_to_str(entry_tag->GetValue()).c_str());

		// Light
		if (entry_light->GetValue() != _T(""))
			items[a]->light = atoi(wx_to_str(entry_light->GetValue()).c_str());
	}

	map.change_level(MC_SECTORS|MC_SSECTS);
}

BEGIN_EVENT_TABLE(SectorEditor, wxDialog)
	EVT_BUTTON(SE_BTN_CHANGESPECIAL, SectorEditor::btn_changespecial_clicked)
	EVT_BUTTON(SE_BTN_FINDTAG, SectorEditor::btn_findtag_clicked)
	EVT_TEXT(SE_ENTRY_FTEX, SectorEditor::entry_ftex_changed)
	EVT_TEXT(SE_ENTRY_CTEX, SectorEditor::entry_ctex_changed)
	EVT_TEXT(SE_ENTRY_SPECIAL, SectorEditor::entry_special_changed)
END_EVENT_TABLE()

void SectorEditor::btn_changespecial_clicked(wxCommandEvent &event)
{
	STypeSelectDialog sd(atoi(wx_to_str(entry_special->GetValue()).c_str()));
	if (sd.ShowModal() == wxID_OK)
		entry_special->SetValue(wxString::Format(_T("%d"), sd.get_value()));
}

void SectorEditor::btn_findtag_clicked(wxCommandEvent &event)
{
	WORD tag = get_free_tag();
	entry_tag->SetValue(wxString::Format(_T("%d"), tag));
}

void SectorEditor::entry_ftex_changed(wxCommandEvent &event)
{
	string tex = wx_to_str(entry_ftex->GetValue().Upper());
	tbox_ftex->set_texture(tex, 2);
}

void SectorEditor::entry_ctex_changed(wxCommandEvent &event)
{
	string tex = wx_to_str(entry_ftex->GetValue().Upper());
	tbox_ftex->set_texture(tex, 2);
}

void SectorEditor::entry_special_changed(wxCommandEvent &event)
{
	if (entry_special->GetValue() == _T(""))
		label_special->SetLabel(_T(""));
	else
	{
		sectortype_t *st = get_sector_type(atoi(wx_to_str(entry_special->GetValue()).c_str()));
		label_special->SetLabel(str_to_wx(st->name));
	}
}
